$TITLE MaxFlowStartPoint OA4202 2010
$INLINECOM { }
 OPTIONS
   SOLPRINT =     OFF,
   DECIMALS =       1,
   LIMCOL   =      10,
   LIMROW   =      10,
   RESLIM   =      60, {max seconds}
   ITERLIM  =99999999, {max iterations}
   LP       =     CPLEX,
   MIP      =     CPLEX
 ;


$offlisting

$ontext
============================================================================
With the following few lines of code, we are reading in the network specification.
- We read in the set of nodes, from the nodes.csv file
- We read in the set of all arcs, from the arcs_set.csv file
- We define the subset of arcs which we can interdict
- Finally, we read in the arc data, from the arcs_data.csv file
============================================================================
$offtext


SET
    n   nodes  /
$INCLUDE  sample_nodes.csv
 /;

alias(n,i,j)   {(i,j) is for arc distance and flow}

SET
   arcs(i,j)         arclist  /
$ONDELIM
$INCLUDE  sample_arcs_set.csv
$OFFDELIM
 /;

SET
   attackarcs(i,j)         the arcs on which we can place interdictions  /
$ONDELIM
$INCLUDE  sample_attack_arcs_set.csv
$OFFDELIM
 /;


table arcdata(i,j,*)
$ONDELIM
$INCLUDE  sample_arcs_data.csv
$OFFDELIM
;

$ontext
============================================================================
Lets define the primal max flow LP.

Our definition includes the parameters xbar(i,j).  We'll use them later, to
re-solve for the max flow, given an interdiction plan X.  For now, we'll just
set the xbar(i,j) to zero.
============================================================================
$offtext

 PARAMETERS
   xbar(i,j)
 ;

 LOOP(arcs(i,j),
   xbar(i,j)=0 ;
 );

 VARIABLE
   Zprimal           objective function value
   V                 the maximum flow variable
 ;

 POSITIVE VARIABLES
   Y(i,j)            flow on arc i j
 ;

 EQUATIONS
   OBJECTIVE
   NETFLOW(n)
   CAPACITY(i,j)
 ;

 OBJECTIVE..
   Zprimal =e=  V - 1.1*SUM(attackarcs(i,j), xbar(i,j)*Y(i,j))
;

 NETFLOW(j)..
   SUM(arcs(i,j),Y(i,j)) - SUM(arcs(j,i),Y(j,i))      {+ flow in - flow out}
     =e= 0 + (-V)$diag(j,'Start')   {start node}
           + (V)$diag(j,'End')   {end node}
 ;
 
 CAPACITY(arcs(i,j))..
   Y(i,j) =l= arcdata(i,j,'Capacity')
 ;

 MODEL MaxFlow
 /
   OBJECTIVE
   NETFLOW
   CAPACITY
 /;

$ontext
============================================================================
Now, lets solve for the max flow when there are no interdictions.
============================================================================
$offtext

 SOLVE MaxFlow USING LP MAXIMIZING Zprimal;       { find the uninterdicted max flow }

$ontext
============================================================================
Lets print the uninterdicted max flow solution to a file called modelOutput.out
============================================================================
$offtext


 file out /modelOutput.out/;
 put out;

 put '****' / ;
 put 'solve MaxFlow with no interdictions' / ;
 LOOP(arcs(i,j)$(Y.l(i,j)>0),
   put '   flow on arc ',i.tl,' -> ', j.tl, ' is ', Y.l(i,j) / ;
 );
 put 'total flow =',Zprimal.l:10:1 / ;


$ontext
============================================================================
Now, lets define the MIP for computing interdictions.

We'll introduce a scalar, 'nAttacks', that specifies the number of
interdictions to compute.

We want to minimize the max flow.  So, the problem we want to solve is a
'min max' problem.  To put this into the solver, we'll take the dual of the
inner problem, so the problem becomes a 'min min'.  That is why in the
following code, you'll see the dual of the max flow problem.

The binary variables X(i,j), will specify whether we attack arc i j.
============================================================================
$offtext


SCALARS nAttacks;

nAttacks = 1;

 VARIABLES
    Zdual
 ;

 VARIABLES
   Rho(n)              dual variables for the netflow constraints
 ;

 POSITIVE VARIABLES
   Pi(i,j)              dual variables for the capacity constraints
 ;

 BINARY VARIABLES
   X(i,j)            should we attack arc i j
 ;

 EQUATIONS
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS(i,j)
   EXTRA_V_CONSTRAINT
   ATTACK_LIMIT
 ;

 DUAL_OBJECTIVE..
   Zdual =e= SUM(arcs(i,j), arcdata(i,j,'Capacity')*Pi(i,j))
 ;

 DUAL_CONSTRAINTS(arcs(i,j))..
   Rho(j)-Rho(i)+Pi(i,j) =g= 0 - 1.1*X(i,j)$(attackarcs(i,j))
 ;

 EXTRA_V_CONSTRAINT..
   Rho('Start') - Rho('End') =e= 1 
 ;

 ATTACK_LIMIT..
   SUM(attackarcs(i,j),X(i,j)) =l= nAttacks
 ;
 
 MODEL InterdictionMIP
 /
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS
   EXTRA_V_CONSTRAINT
   ATTACK_LIMIT
 /;

$ontext
============================================================================
Now, lets solve for the best interdiction plan.
============================================================================
$offtext

 InterdictionMIP.optfile = 1;

 SOLVE InterdictionMIP USING MIP MINIMIZING Zdual;       { Find the best interdiction plan }

$ontext
============================================================================
Lets print the interdiction plan into the output file we had opened, modelOutput.out
============================================================================
$offtext

display X.l;

 put '****' / ;
 put 'interdiction plan with ', nAttacks, ' nAttacks:' / ;
 LOOP(arcs(i,j)$(X.l(i,j)>0),
   put '   attack arc: ',i.tl,' -> ', j.tl / ;
 );
 put 'flow with interdictions in place =',Zdual.l:10:1 / ;

$ontext
============================================================================
Finally, now that we have computed attack locations, lets re-solve for 
the operator's new best route.
============================================================================
$offtext

 LOOP(arcs(i,j),                        { transfer the interdiction solution to constant parameter }
   xbar(i,j)=X.l(i,j) ;
 );

 SOLVE MaxFlow USING LP MAXIMIZING Zprimal;       { find shortest escape route the interdictions in place }

 put '****' / ;
 put 'solve MaxFlow with interdictions in place' / ;
 put 'operator best response:' / ;
 LOOP(arcs(i,j)$(Y.l(i,j)>0),
   put '   flow on arc ',i.tl,' -> ', j.tl, ' is ', Y.l(i,j)/ ;
 );
 put 'total flow =',Zprimal.l:10:1 / ;
